/* -*- fundamental -*- */
%{
#include "com.h"
#include "symtab.h"
#include "y.tab.h"

extern char *file;
static int line = 1;
static char strbuf[1024];
static char *strptr;
%}

%x comment
%x string
%x verbatim

%%
struct			return STRUCT;
typedef			return TYPEDEF;
unsigned		return UNSIGNED;
notypedef		return NOTHDR;
generate		return GENERATE;

	/* Purely C keywords */
static			return STATIC;
extern			return EXTERN;
(__)?const		return CONST;
(__)?volatile		return VOLATILE;
"..."			return ELLIPSIS;

[_a-zA-Z][_a-zA-Z0-9]*	{
				Var *v = lookup(global, yytext);
				if (v != NULL && v->type->class == Typedef)
				{
					yylval.type = v->type->i.typedef_i;
					return TYPE;
				}
				yylval.name = strdup(yytext);
				return IDENT;
			}
[0-9]+			{ yylval.num = atoi(yytext); return NUM; }
[+-/{}[\]*;,()]		return *yytext;
"/*"			BEGIN(comment);
^#\ [0-9]+\ \"[a-zA-Z_.0-9/]+\".*$	{
				char *cp;

				line = strtol(&yytext[2], &cp, 10);
				free(file);
				file = cp+2;
				for(cp = file; *cp && *cp != '"'; cp++)
					;
				*cp = 0;
				file = strdup(file);
			}
\"			strptr = strbuf; BEGIN(string);
^#.*$			;
[ \t]			;
\n			line++;
^"%%"			BEGIN(verbatim);
.			fprintf(stderr, "Uncaught char '%c'\n", *yytext);

<comment>[^*\n]+	/* eat up anything up to '*' */
<comment>"*"+[^*/\n]*	/* eat up rows of '*' not followed by '/' */
<comment>\n		line++;
<comment>"*"+"/"	BEGIN(INITIAL);

<string>\\.		*strptr++ = yytext[1];
<string>\"		*strptr=0; yylval.name = strdup(strbuf); BEGIN(INITIAL); return STRING;
<string>[^\\\"]+	{char *cp = yytext; while(*cp) *strptr++ = *cp++; }

<verbatim>^"%%"		BEGIN(INITIAL);
<verbatim>^%.*		fputs(&yytext[1], stdout);
<verbatim>.*		fputs(yytext, stdout);
<verbatim>\n		line++; putchar('\n');

%%

void
yyerror(char *msg)
{
	fprintf(stderr, "%s:%d: %s\n", file, line, msg);
}

