/*
 * Symbol table structures
 */

typedef struct var_t Var;
typedef struct type_t Type;

struct type_t
{
	enum
	{
		Struct,
		Typedef,
		Scalar,
		Array,
		CArray,
		Pointer,
		Void
	} class;
	union
	{
		Var 	*struct_i;
		Type	*typedef_i;
		struct
		{
			enum sc_size
			{
				Char,
				Short,
				Long,
				Float,
				Double
			} size;
			int	Unsigned;
		} scalar_i;
		Type	*array_i;
		Type	*pointer_i;
		struct
		{
			Type *type;
			unsigned size;
		} carray_i;
	} i;
	char *alias;
	int	anon:1;		/* Had no defined name */
	int	nocode:1;	/* No code to be generated */
	int	nohdr:1;	/* No header generated */
	int	demand:1;	/* Generate code as needed */
	int	norep:1;	/* No serial representation */
	int	coded:3;	/* En/Decode func generated */
};

struct var_t
{
	char	*name;
	Type	*type;
	Var	*next;
	Var	*prev;
};

extern Var *global;	/* Global symbol table */
extern Var *structures;	/* Global structure table */

void init_symtab(void);

Var *lookup(Var *, const char *);
Var *add(Var *, Var *);
Var *addl(Var *, Var *);
Var *preadd(Var *, Var *);

Type *mktype(void);
Type *mkstruct(Var *);
Type *mkpointer(Type *);
Type *mkarray(Type *);
Type *mkcarray(Type *, unsigned int);
Type *mkscalar(enum sc_size, int);
Type *mkvoid(void);

Var *mkvar(Type *, char *);
Var *mktypedef(Type *, char *);
Var *mkincl(const char *);

Var *name_types(Var *);

/* Standard types */
extern Type *Ulong_type;
extern Type *Ushort_type;
extern Type *Uchar_type;
extern Type *long_type;
extern Type *short_type;
extern Type *char_type;
extern Type *void_type;
extern Type *float_type;
extern Type *double_type;
