#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <syscall.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dirent.h>

extern "C"
{
#include <linux/fs.h>
#include <linux/userfs_fs.h>
}

#include "itab.h"
#include "uio.h"

extern itab itable;
extern void *xmalloc(size_t);

static _syscall3(int, readdir, unsigned int, fd, struct dirent *, de, int, count);

int do_readdir(unsigned char *buf, upp_repl *repl)
{
	int n;
	int ret;
	int dfd;
	struct dirent de;
	struct stat st;
	char name[1024], *ep;
	ient *ino;
	upp_readdir_s rcv;
	upp_readdir_r snd;

	snd.name.nelem = 0;

	decode_upp_readdir_s(&rcv, buf);
	
	DB(printf("d=%x off=%d ", rcv.dir.handle, rcv.off));
	if ((ino = itable.get_inode(rcv.dir.handle)) == NULL)
	{
		fprintf(stderr, "%s:%d: bad dir inode/handle %x\n",
			__FILE__, __LINE__, rcv.dir.handle);
		return EINVAL;
	}

	snd.off = 0;
	if (!ino->isopen() || ino->goo != rcv.off)
	{
		DB(putchar('C'));
		if ((dfd = ino->open()) == -1)
		{
			ret = errno;
			FAIL("dir open 1");
			goto bad;
		}

		lseek(dfd, 0, SEEK_SET);

		for(n = 0; n < rcv.off; n++)
			if ((ret = readdir(dfd, &de, 1)) <= 0)
			{
				if (ret == -1)
					ret = errno;
				goto bad;
			}
	}
	else
	{
		DB(putchar('O'));
		if ((dfd = ino->open()) == -1)
		{
			ret = errno;
			FAIL("dir open 2");
			goto bad;
		}
	}

	if ((ret = readdir(dfd, &de, 1)) <= 0)
	{
		if (ret == -1)
			ret = errno;
		goto bad;
	}

	snd.name.nelem = de.d_reclen;
	snd.name.elems = de.d_name;

	ep = ino->catname(name);
	*ep++ = '/';
	strcpy(ep, de.d_name);

	if (lstat(name, &st) == -1)
		FAIL("lstat");
	
	snd.file.handle = itable.genhand(st.st_dev, st.st_ino);
	snd.off = 1;
	ino->goo = rcv.off + 1;

	DB(printf("n=%*s nl=%d h=%x ",
		snd.name.nelem, snd.name.elems, snd.name.nelem, snd.file.handle));
	ret = 0;
 bad:
	repl->size = encode_upp_readdir_r(&snd, buf) - buf;

	ino->close();
	return ret;
}
