#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#include <linux/fs.h>
#include <linux/userfs_fs.h>

#include "uio.h"

static int myread(int fd, void *buf, int sz)
{
	int	rd = 0;
	int	ret;
	
	while(sz > 0)
	{
		ret = read(fd, buf, sz);

		if (ret == -1)
		{
			if (errno == EINTR)
				continue;

			perror("myread");
			exit(1);
		}

		if (ret == 0)
			break;
				
		sz -= ret;
		buf += ret;
		rd += ret;
	}

	return rd;
}

static int mywrite(int fd, const void *buf, int sz)
{
	int	wr = 0;
	int	ret;
	
	while(sz > 0)
	{
		ret = write(fd, buf, sz);

		if (ret == -1)
		{
			if (errno == EINTR)
				continue;
			FAIL("write");
			exit(1);
		}

		if (ret == 0)
		{
			fprintf(stderr, "write wrote 0 bytes\n");
			break;
		}

		sz -= ret;
		buf += ret;
		wr += ret;
	}

	return wr;
}

int readpkt(int fd, up_preamble *pre, unsigned char *buf)
{
	int presz = sizeof_up_preamble(pre);
	unsigned int rd;
	unsigned char pbuf[presz];

	memset(pbuf, 0, sizeof(pbuf));

	if ((rd = myread(fd, pbuf, presz)) != presz)
	{
		if (rd == 0)
			return 0;
		if (rd > 0)
			fprintf(stderr, "readpkt, incomplete read (%d, not %d)\n",
				rd, presz);
		perror("readpkt, header read");
		return -1;
	}

	decode_up_preamble(pre, pbuf);

	if (pre->isreq != UP_ENQ && pre->isreq != UP_REQ)
		fprintf(stderr, "pre->isreq = %d\n", pre->isreq);

	if (pre->version != UP_VERSION)
	{
		fprintf(stderr, "Version mismatch: us=%d them=%d\n",
			UP_VERSION, pre->version);
		return -1;
	}


	if (pre->size != 0)
	{
		if (myread(fd, buf, pre->size) != pre->size)
		{
			perror("readpkt, body read");
			return -1;
		}
	}

	return 1;
}

int writepkt(int fd, upp_repl *repl, const unsigned char *buf)
{
	int replsz = sizeof_upp_repl(repl);
	unsigned char pbuf[replsz];
	unsigned int wr;
	unsigned char *p;

	assert(repl->version == UP_VERSION);
	assert(repl->isreq == UP_REPL);

	p = encode_upp_repl(repl, pbuf);

	assert(replsz == p - pbuf);

	if ((wr = mywrite(fd, pbuf, replsz)) != replsz)
	{
		perror("writepkt, header write");
		return -1;
	}

	if (repl->size > 0 && repl->errno == 0)
	{
		if (mywrite(fd, buf, repl->size) != repl->size)
		{
			perror("writepkt, body write");
			return -1;
		}
	}

	return 1;
}

