// -*- c++ -*-
//
// User level filesystem class
//
// Jeremy Fitzhardinge, 1993

#include "Filesystem.h"
#include <unistd.h>

Filesystem::Filesystem(int to, int from)
{
	tokern = to;
	fromkern = from;
}

Filesystem::~Filesystem()
{
	close(tokern);
	close(fromkern);
}

static void genrepl(const up_preamble &pre, upp_repl repl)
{
	repl.size = 0;
	repl.errno = 0;
	repl.isreq = UP_REPL;
	repl.seq = pre.seq;
	repl.op = pre.op;
}

int
Filesystem::getreq(up_preamble &pre, Data *&data)
{
	Data prebuf[sizeof_up_preamble(&pre)];
	int ret;
	
	data = NULL;
	
	if ((ret = fsrecv(prebuf, sizeof(prebuf))) != sizeof(prebuf))
		return ret;

	decode_up_preamble(&pre, prebuf);
	
	if (pre.isreq == UP_REQ)
	{
		if (pre.size > 0)
			data = new Data[pre.size];
		
		if ((ret = fsrecv(data, sizeof(data))) != 0)
		{
			delete data;
			data = NULL;
			return ret;
		}
	}
	
	return 0;
}

int
Filesystem::DoOp()
{
	up_preamble pre;
	upp_repl repl;
	Data *data;
	int err = 0;
	int ret;
	
	ret = getreq(pre, data);

	if (ret != 0)
		return 0;

	switch(pre.op)
	{
#define DO(op)	case up_##op: \		// Super operations
			err = fs_##op(pre.isreq == UP_ENQ, repl, data); \
			break
	
		DO(mount);
		DO(iread);
		DO(iwrite);
		DO(iput);
		DO(umount);
		DO(statfs);
#undef DO
		
	default:
		err = ENOSYS;
		break;
	}

	repl.errno = err;

	if (errno != 0)
		assert(repl.size == 0);
	if (pre.isreq == UP_ENQ)
		assert(repl.size == 0);
	
	return 1;
}




